#!/usr/bin/env python

"""
Confluence Wiki syntax parsing.

Copyright (C) 2012 Paul Boddie <paul@boddie.org.uk>

This software is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License as
published by the Free Software Foundation; either version 2 of
the License, or (at your option) any later version.

This software is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public
License along with this library; see the file LICENCE.txt
If not, write to the Free Software Foundation, Inc.,
51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA

--------

The basic procedure is as follows:

 1. Wiki pages are first split up into regions.
 2. Then, within these regions, the text is split into blocks.
    1. First, lists are identified.
    2. Additionally, other block-like elements are identified.
 3. Each block is then parsed.
"""

import re

# Section extraction.

sections_regexp_str = r"(?<!{){(?P<type>[^{}\n:]+)(:[^}\n]+)?}.*?{(?P=type)}"
sections_regexp = re.compile(sections_regexp_str, re.DOTALL | re.MULTILINE)

def get_regions(s):

    """
    Return a list of regions from 's'. Each region is specified using a tuple of
    the form (type, text).
    """

    last = 0
    regions = []
    for match in sections_regexp.finditer(s):
        start, end = match.span()
        regions.append((None, s[last:start]))
        regions.append(get_section_details(s[start:end]))
        last = end
    regions.append((None, s[last:]))
    return regions

# Section inspection.

section_regexp_str = r"{(?P<sectiontype>[^\n:]*?)(?::(?P<options>.*?))?}(?P<section>.*){(?P=sectiontype)}"
section_regexp = re.compile(section_regexp_str, re.DOTALL | re.MULTILINE)

def get_section_details(s):

    "Return the details of a section 's' in the form (type, text)."

    match = section_regexp.match(s)
    if match:
        return (match.group("sectiontype"), match.group("options")), match.group("section")
    else:
        return None, s

# Heading, table and list extraction.

list_regexp_str = r"^(?P<listtype>[*#-])[*#-]*.*\n((?P=listtype).*(?:\n|$))*"
table_regexp_str = r"^((?P<celltype>[|]{1,2})(.+?(?P=celltype))+(\n|$))+"
blocktext_regexp_str = r"^(?P<type>h\d|bq)\.\s+(?P<text>.*)$"

blockelement_regexp = re.compile(
    "(" + list_regexp_str + ")"
    "|"
    "(" + table_regexp_str + ")"
    "|"
    "(" + blocktext_regexp_str + ")",
    re.MULTILINE
    )

def get_block_elements(s):

    """
    Extract headings, tables and lists from the given string 's'.
    """

    last = 0
    blocks = []
    for match in blockelement_regexp.finditer(s):
        start, end = match.span()
        matchtype = match.group("listtype") and "list" or match.group("celltype") and "table" or match.group("type")
        blocks.append((None, s[last:start]))
        blocks.append((matchtype, match.group("text") or s[start:end]))
        last = end
    blocks.append((None, s[last:]))
    return blocks

# Block extraction.

block_regexp_str = r"^(?:\s*\n)+"
block_regexp = re.compile(block_regexp_str, re.MULTILINE)

def get_basic_blocks(s):

    """
    Return blocks from the given string 's' by splitting the text on blank lines
    and eliminating those lines.
    """

    return [b for b in block_regexp.split(s) if b.strip()]

# Block inspection.

def get_blocks(s):

    """
    Return blocks from the given string 's', inspecting the basic blocks and
    generating additional block-level text where appropriate.
    """

    blocks = []

    for blocktype, blocktext in get_block_elements(s):

        # Collect heading, list and table blocks.

        if blocktype is not None:
            blocks.append((blocktype, blocktext))

        # Attempt to find new subblocks in other regions.

        else:
            for block in get_basic_blocks(blocktext):
                blocks.append((None, block))

    return blocks

# List item inspection.

listitem_regexp_str = r"^(?P<marker>[*#-])+\s*(?P<text>.*)$"
listitem_regexp = re.compile(listitem_regexp_str, re.MULTILINE)

def get_list_items(text):

    "Return a list of (marker, text) tuples for the given list 'text'."

    items = []

    for match in listitem_regexp.finditer(text):
        items.append((match.group("marker"), match.group("text")))

    return items

# Table row inspection.

link_regexp_str = r"[[](?P<linktext>.*?)]"
image_regexp_str = r"!(?P<imagetext>.*?)!"
cellsep_regexp_str = r"(?P<celltype>[|]{1,2})"

content_regexp_str = (
    "(" + link_regexp_str + ")"
    "|"
    "(" + image_regexp_str + ")"
    )

table_content_regexp_str = (
    content_regexp_str +
    "|"
    "(" + cellsep_regexp_str + ")"
    )

content_regexp = re.compile(content_regexp_str)
table_content_regexp = re.compile(table_content_regexp_str)

def translate_content_match(match):

    "Translate the content described by the given 'match', returning a string."

    if match.group("linktext"):
        parts = match.group("linktext").split("|")

        # NOTE: Proper detection of external links required.

        if len(parts) > 1 and parts[1].startswith("http"):
            prefix = ""
        elif parts[0].startswith("#"):
            prefix = ""
        elif parts[0].startswith("^"):
            prefix = "attachment:"
        else:
            prefix = "../"

        if len(parts) == 1:
            return "[[%s%s]]" % (prefix, parts[0])
        elif len(parts) == 2:
            return "[[%s%s|%s]]" % (prefix, parts[1], parts[0])
        else:
            return "[[%s%s|%s|title=%s]]" % (prefix, parts[1], parts[0], parts[2])

    elif match.group("imagetext"):
        parts = match.group("imagetext").split("|")

        # NOTE: Proper detection of external links required.

        if parts[0].startswith("http"):
            prefix = ""
        else:
            prefix = "attachment:"

        # NOTE: Proper options conversion required.

        if len(parts) == 1:
            return "{{%s%s}}" % (prefix, parts[0])
        else:
            return "{{%s%s|%s}}" % (prefix, parts[1], parts[0])

    else:
        return match.group()

def get_table_rows(text):

    "Return a list of (cellsep, columns) tuples for the given table 'text'."

    rows = []

    for line in text.split("\n"):
        cellsep = None
        columns = [""]
        last = 0
        for match in table_content_regexp.finditer(line):
            start, end = match.span()
            columns[-1] += line[last:start]

            if match.group("celltype"):
                if cellsep is None:
                    cellsep = match.group("celltype")
                columns.append("")
            else:
                columns[-1] += match.group()

            last = end

        columns[-1] += line[last:]

        if cellsep:
            rows.append((cellsep, columns[1:-1]))

    return rows

def translate_content(text):

    "Return a translation of the given 'text'."

    parts = []

    last = 0
    for match in content_regexp.finditer(text):
        start, end = match.span()
        parts.append(text[last:start])
        parts.append(translate_content_match(match))
        last = end

    parts.append(text[last:])
    return "".join(parts)

# Translation helpers.

blocktypes = {
    "h1" : "= %s =",
    "h2" : "== %s ==",
    "h3" : "=== %s ===",
    "h4" : "==== %s ====",
    "h5" : "===== %s =====",
    "h6" : "====== %s ======",
    "bq" : "{{{%s}}}",
    }

markers = {
    "*" : "*",
    "#" : "1.",
    "-" : "*",
    }

def translate_marker(marker):

    "Translate the given 'marker' to a suitable Moin representation."

    return " " * len(marker) + markers[marker[-1]]

cellseps = {
    "|" : "||",
    "||" : "||",
    }

cellextra = {
    "|" : "",
    "||" : "'''",
    }

def translate_cellsep(cellsep):

    "Translate the given 'cellsep' to a suitable Moin representation."

    return cellseps[cellsep]

def translate_cell(cellsep, text):

    "Using 'cellsep', translate the cell 'text'."

    return cellextra[cellsep] + translate_content(text) + cellextra[cellsep]

sectiontypes = {
    "code" : "",
    "noformat" : "",
    "quote" : "",
    "info" : "wiki important",
    "note" : "wiki caution",
    "tip" : "wiki tip",
    "warning" : "wiki warning",
    }

# General parsing.

def parse(s, out):

    "Parse the content in the string 's', writing a translation to 'out'."

    for type, text in get_regions(s):

        # Handle list, heading, blockquote or anonymous blocks.

        if type is None:
            for blocktype, blocktext in get_blocks(text):

                # Translate headings and blockquotes.

                if blocktypes.has_key(blocktype):
                    print >>out, blocktypes[blocktype] % blocktext

                # Translate list items.

                elif blocktype == "list":
                    for listmarker, listitem in get_list_items(blocktext):
                        print >>out, "%s %s" % (translate_marker(listmarker), translate_content(listitem))

                # Translate table items.

                elif blocktype == "table":
                    for cellsep, columns in get_table_rows(blocktext):
                        moinsep = translate_cellsep(cellsep)
                        print >>out, moinsep + moinsep.join([translate_cell(cellsep, column) for column in columns]) + moinsep

                # Handle anonymous blocks.

                else:
                    print >>out, translate_content(blocktext.rstrip())

                print >>out

        # Handle sections.

        else:
            sectiontype, options = type

            # Direct translations of sections.

            mointype = sectiontypes.get(sectiontype)
            if mointype:
                print >>out, "{{{#!%s" % mointype
                if options:
                    print >>out, "##", options
            else:
                print >>out, "{{{",
            print >>out, translate_content(text),
            print >>out, "}}}"
            print >>out

if __name__ == "__main__":
    import sys

    s = sys.stdin.read()
    parse(s, sys.stdout)

# vim: tabstop=4 expandtab shiftwidth=4
